/*
 *	Floppy-Disk Mounter 1.3 (22-12-96)
 *	(c) 1996 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

/*
   HISTORY:

   07-03-95 - version 1.1:

   - mount point directory correctly removed if mount failed
   - DOS floppies have the noexec bit set
   - better message informing about R/W/RW access type
   - break problems fixed
   - when device already mounted with specific format, correct msg is printed 

   12-04-96 - version 1.2:

   - group "disk" renamed to "floppy"
   - umask problems fixed

   22-12-96 - version 1.3:

   - code cleanup
   - re-indented
   - "vfat" added to the list of allowed filesystems
   - "-d" switch (no format autodetection) implemented
   - return codes fixed
 */

/* Starting with some includes ... */

#define __KERNEL__
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <linux/fd.h>
#include <linux/ext2_fs.h>
#include <linux/msdos_fs.h>
#include <linux/major.h>
#include <unistd.h>
#include <signal.h>

gid_t floppy_gid;						/* GID of 'floppy' group */
int unmount;							/* '-u' switch specified */
int autodetect = 1;						/* Autodetection of format enabled */
int exit_code;							/* Exit code */
int kill_dir_on_exit;					/* Remove the mount point before exitting */

void err (char *);

/* String comparison */

#define strlcmp(a,b,l) (strlen(a) != l || strncmp(a,b,l))

/* Some variables for mtab mechanism */

int lockfd;								/* File descriptor of /etc/mtab~ */
int mtab_open;							/* mtab mechanism initialized */
int mtab_written;						/* mtab changed */
char *lockname = "/etc/mtab~";			/* paths */
char *mtabname = "/etc/mtab";

/*** EXIT ROUTINES ***/

void remove_mp (void);
void end_mtab (void);
void forbid_breaks (void);

void 
myexit (int i)
{
  forbid_breaks ();
  if (mtab_open)
    end_mtab ();
  if (kill_dir_on_exit)
    remove_mp ();
  exit (i);
}

/* Exit with error message */

void 
err (char *c)
{
  fprintf (stderr, "%s !\n", c);
  myexit (1);
}

/*** BREAKS ***/

/* Forbid all breaks */

void 
forbid_breaks (void)
{
  signal (SIGINT, SignalIgnore);
  signal (SIGQUIT, SignalIgnore);
}

/* Break handler */

void 
handle_break (void)
{
  forbid_breaks ();
  err ("Break");
}

/* Block ordinary breaks and replace them by our custom stuff */

void 
set_breaks (void)
{
  signal (SIGINT, (SignalHandler) handle_break);
  signal (SIGQUIT, (SignalHandler) handle_break);
}

/*** MANIPULATION WITH /etc/mtab ***/

struct mtent
{
  struct mtent *next;
  char ent[1];
};

struct mtent *firstmt, **lastmt;

/* Allocate memory and exit if failed */

void *
xalloc (int l)
{
  void *m = malloc (l);
  if (!m)
    err ("Insufficient memory");
  return m;
}

/* Add entry into mtab in memory -- internal */

void 
add_mtab (char *entry)
{
  struct mtent *r;
  r = xalloc (sizeof (struct mtent) + strlen (entry));
  *lastmt = r;
  lastmt = &r->next;
  *lastmt = NULL;
  strcpy (r->ent, entry);
}

/* Initialize mtab mechanism */

void 
start_mtab (void)
{
  char lin[256];
  FILE *f;
  lockfd = open (lockname, O_WRONLY | O_EXCL | O_CREAT, 0644);
  if (lockfd < 0)
    err ("Cannot create lock file /etc/mtab~");
  lastmt = &firstmt;
  if (!(f = fopen (mtabname, "r")))
    err ("Cannot access /etc/mtab");
  for (;;)
    {
      fgets (lin, 256, f);
      if (feof (f))
		break;
      add_mtab (lin);
    }
  fclose (f);
  mtab_open = 1;
  mtab_written = 0;
}

/* Flush mtab buffers and finish all operations */

void 
end_mtab (void)
{
  FILE *f;
  if (mtab_written)
    {
      if (!(f = fdopen (lockfd, "w")))
		err ("Error writing /etc/mtab~");
      while (firstmt)
		{
		  if (fputs (firstmt->ent, f) == EOF)
			err ("Write error on /etc/mtab~");
		  firstmt = firstmt->next;
		}
      fclose (f);
      unlink (mtabname);
      if (rename (lockname, mtabname))
	err ("Cannot rename /etc/mtab~ to /etc/mtab");
    }
  else
    {
      close (lockfd);
      unlink (lockname);
    }
}

/* Get start and length of n-th field in mtab entry */

int 
get_mt_field (struct mtent *e, int n, char **start, int *len)
{
  char *x = NULL, *c = e->ent;
  while (n > 0)
    {
      while (*c == ' ' || *c == '\t')
	c++;
      x = c;
      if (!*c)
	return 0;
      while (*c && *c != ' ' && *c != '\t')
	c++;
      n--;
    }
  *start = x;
  *len = c - x;
  return 1;
}

/* Update mtab entry */

void 
update_mtab (char *device, char *mpoint, char *fs, char *options)
{
  struct mtent **r, *z;
  char buf[256];
  for (r = &firstmt; (z = *r); r = &z->next)
    {
      char *l;
      int n;
      if (get_mt_field (z, 1, &l, &n) && !strlcmp (device, l, n))
	{
	  *r = z->next;
	  if (lastmt == &z->next)
	    lastmt = r;
	  free (z);
	  mtab_written = 1;
	  break;
	}
    }
  if (mpoint)
    {
      sprintf (buf, "%s %s %s %s 0 0\n", device, mpoint, fs, options);
      add_mtab (buf);
      mtab_written = 1;
    }
}

/* Find device mounted to given mount point */

int 
find_device_by_mp (char *dest, char *mp)
{
  struct mtent *r;
  char *c;
  int l;
  for (r = firstmt; r; r = r->next)
    if (get_mt_field (r, 2, &c, &l) && !strlcmp (mp, c, l))
      {
	get_mt_field (r, 1, &c, &l);
	dest[l] = 0;
	strncpy (dest, c, l);
	return 1;
      }
  return 0;
}

/* Passwd and group structures */

struct passwd *user_pas;
struct group *floppy_grp, *device_grp;

/* Is given user in given group? */

int 
ingroup (struct group *g, struct passwd *p)
{
  char **c;
  if (p->pw_gid == g->gr_gid)
    return 1;			/* Primary group */
  for (c = g->gr_mem; *c; c++)
    if (!strcmp (*c, p->pw_name))
      return 1;
  return 0;
}

/* Check access rights for mounting floppies */

void 
checkuser (void)
{
  umask (0022);
  if (!(floppy_grp = getgrnam ("floppy")))
    err ("Group 'floppy' not defined");
  floppy_gid = floppy_grp->gr_gid;
  if (!(user_pas = getpwuid (getuid ())))
    err ("Cannot identify current user");
  if (!getuid ())
    return;			/* We are logged in as root */
  if (!ingroup (floppy_grp, user_pas))
    err ("You are not allowed to do this");
  if (geteuid ())
    err ("flop must be setuid root");
}

char dev[256], fs[256];		/* Device name and filesystem type */
struct stat rootstat;		/* stat record for root */
dev_t floppy_dev;		/* Device code of current floppy */

/* Construct device name from floppy drive number */

void 
devnum (int i)
{
  sprintf (dev, "/dev/fd%d", i);
}

/* Check access to given device */

int 
check_device (void)
{
  struct stat st;
  if (!stat (dev, &st))
    {
      floppy_dev = st.st_rdev;
      if ((st.st_mode & S_IFMT) == S_IFBLK
		  && st.st_gid == floppy_gid
		  && (st.st_mode & 040)
		  )
		return 1;
    }
  return 0;
}

/* Return floppy disk unit number from device number */

int 
fdu (dev_t dev)
{
  int j = MINOR (dev);
  return (j & 3) + ((j & 0x80) >> 4);
}

/* Determine mount point for given device */

char mp[256];
void 
mount_point (void)
{
  if (strncmp (dev, "/dev/fd", 7) || !dev[7])
    strcpy (mp, "/mnt");
  else
    sprintf (mp, "/fd%c", dev[7]);
}

/* Return what is mounted to given mount point:
   0 = nothing
   1 = something substantially different from current device
   2 = current device
   3 = unit same as current device
 */

int 
is_mounted (char *mp)
{
  struct stat st;
  if (stat (mp, &st))
    return 0;
  if (st.st_dev == rootstat.st_dev)
    return 0;
  if (st.st_dev == floppy_dev)
    return 2;
  if (MAJOR (st.st_dev) == FLOPPY_MAJOR && fdu (st.st_dev) == fdu (floppy_dev))
    return 3;
  return 1;
}

/* Correct device according to known number of sectors */

void 
correct_device (int *fd, int sectors)
{
  struct floppy_drive_params par;
  char devold[256];
  unsigned char mask, type;
  if (ioctl (*fd, FDGETDRVPRM, &par))
    err ("FDGETDRVPRM ioctl() failed");
  strcpy (devold, dev);
  type = (sectors >= 2048) ? 'H' : 'D';
  switch (par.cmos)
    {
    case 1:
    case 2:
      mask = 0xdf;
      break;
    default:
      mask = 0xff;
    }
  sprintf (dev + 8, "%c%d", type & mask, sectors / 2);
  if (check_device ())
    {
      close (*fd);
      *fd = open (dev, O_RDONLY);
      if (fd < 0)
	err ("unable to open");
      printf ("using %s - ", dev);
    }
  else
    {
      printf ("%d sectors - ", sectors);
      strcpy (dev, devold);
    }
  fflush (stdout);
}

/* Try to autodetect filesystem type */

int 
detect_fs (int *fd)
{
  unsigned char sector[1536];
  struct msdos_boot_sector *b = (struct msdos_boot_sector *) sector;
  struct ext2_super_block *s = (struct ext2_super_block *) (sector + 1024);
  unsigned int sect = 0;
  if (fs[0])
    return 1;
  if (read (*fd, sector, 1536) != 1536)
    return 0;
  if (sector[511] == 0xaa && sector[510] == 0x55
      && b->sector_size[0] == 0 && b->sector_size[1] == 2 && b->fats == 2)
    {
      strcpy (fs, "msdos");
      sect = ((int) b->sectors[1] << 8) | (b->sectors[0]);
      if (!sect)
		sect = b->total_sect;
    }
  else
    {
      strcpy (fs, "ext2");
      if (s->s_magic != EXT2_SUPER_MAGIC || s->s_creator_os != EXT2_OS_LINUX)
		{
		  printf ("suspicious superblock - ");
		  return 0;
		}
      else
		sect = s->s_blocks_count * 2;
    }
  if (sect && autodetect)
    correct_device (fd, sect);
  return 1;
}

/* Remove mount point */

void 
remove_mp (void)
{
  if (strcmp (mp, "/mnt"))
    rmdir (mp);							/* /mnt won't be removed */
  kill_dir_on_exit = 0;
}

/* Get access type string */

char *access_names[] =
{"inaccessibly", "writeonly", "readonly", "read/write"};

char *
get_access (int readonly)
{
  struct stat st;
  uid_t uid;
  int attr;
  if (stat (mp, &st))
    return "very suspicously";
  uid = getuid ();
  if (uid == st.st_uid)
    attr = st.st_mode >> 6;
  else if ((device_grp = getgrgid (st.st_gid)) && ingroup (device_grp, user_pas))
    attr = st.st_mode >> 3;
  else
    attr = st.st_mode;
  if (!uid)
    st.st_mode = 7;						/* root is the almighty one */
  if (readonly)
    attr &= 5;
  return access_names[(attr >> 1) & 3];
}

/* Perform mount/unmount of current device */

void 
do_mount (void)
{
  int fd, i;
  struct floppy_drive_struct status;
  printf ("%s - ", dev);
  fflush (stdout);
  mount_point ();
  if (unmount)
    {
      switch (is_mounted (mp))
		{
		case 0:
		  printf ("not mounted.\n");
		  break;
		case 1:
		  printf ("mount point %s is used by some other device.\n", mp);
		  exit_code = 1;
		  break;
		case 3:
		  if (find_device_by_mp (dev, mp))
			{
			  printf ("using %s - ", dev);
			  fflush (stdout);
			}
		  else
			{
			  printf ("cannot identify format type.\n");
			  break;
			}
		  /* Fall-thru */
		default:
		  if (umount (mp))
			{
			  printf ("unmounting failed (%s).\n", strerror (errno));
			  exit_code = 1;
			}
		  else
			{
			  printf ("unmounted.\n");
			  update_mtab (dev, NULL, NULL, NULL);
			  remove_mp ();
			}
		}
    }
  else
    {
      switch (is_mounted (mp))
		{
		case 1:
		  printf ("mount point %s is busy.\n", mp);
		  exit_code = 1;
		  break;
		case 2:
		case 3:
		  printf ("already mounted on %s.\n", mp);
		  break;
		default:
		  set_breaks ();
		  fd = open (dev, O_RDONLY);
		  forbid_breaks ();
		  if (fd < 0)
			{
			  puts ("not accessible.");
			  exit_code = 1;
			  return;
			}
		  mkdir (mp, 0777);
		  chown (mp, 0, floppy_gid);
		  chmod (mp, 0770);
		  kill_dir_on_exit = 1;
		  set_breaks ();
		  i = detect_fs (&fd);
		  forbid_breaks ();
		  if (!i)
			{
			  printf ("FS detection failed.\n");
			  exit_code = 1;
			}
		  else
			{
			  char *param = NULL, mparam[256];
			  int flg = MS_NOSUID | MS_NODEV;
			  if (ioctl (fd, FDGETDRVSTAT, &status))
				err ("ioctl() failed");
			  if (!strcmp (fs, "msdos"))
				{
				  param = mparam;
				  sprintf (mparam, "gid=%d,umask=002", (int) floppy_gid);
				  flg |= MS_MGC_VAL | MS_NOEXEC;
				}
			  if (!(status.flags & FD_DISK_WRITABLE))
				flg |= MS_RDONLY;
			  if (mount (dev, mp, fs, flg, param))
				{
				  printf ("not mounted (%s).\n", strerror (errno));
				  exit_code = 1;
				}
			  else
				{
				  char opt[256];
				  kill_dir_on_exit = 0;
				  printf ("mounted %s on %s, fs=%s.\n",
						  get_access (!(status.flags & FD_DISK_WRITABLE)),
						  mp, fs);
				  opt[0] = 0;
				  if (param)
					{
					  strcpy (opt, param);
					  strcat (opt, ",");
					}
				  strcat (opt, (status.flags & FD_DISK_WRITABLE) ?
						  "rw" : "ro");
				  update_mtab (dev, mp, fs, opt);
				}
			}
		}
      close (fd);
    }
  if (kill_dir_on_exit)
    remove_mp ();
}

/* Table of allowed filesystem names */

char *allowed_filesys[] =
{"msdos", "minix", "ext2", "vfat", NULL};

/* Analyse arguments given directly */

void 
doarg (char *a)
{
  if (strlen (a) > 255)
    err ("Argument too long");
  fs[0] = 0;
  if (*a >= '0' && *a <= '3' && (!a[1] || a[1] == '='))
    devnum (*a++ - '0');
  else
    {
      char *d = dev;
      if (*a != '/')
		{
		  strcpy (d, "/dev/");
		  d += 5;
		}
      while (*a && *a != '=')
		*d++ = *a++;
      *d = 0;
    }
  if (*a++ && *a)
    {
      char **c = allowed_filesys;
      while (*c)
		if (!strcmp (*c++, a))
		  break;
      if (!*c)
		err ("Wrong filesystem type");
      strcpy (fs, a);
    }
  if (strncmp (dev, "/dev/", 5))
    err ("Wrong device");
  if (!check_device ())
    err ("Device not accessible");
  do_mount ();
}

/* Return last component of path */

char *
truncate_path (char *c)
{
  char *l = c;
  while (*c)
    if (*c++ == '/')
      l = c;
  return l;
}

/* Main program -- parse arguments and call subroutines */

void 
main (int argc, char **argv)
{
  int i;
  if (argc == 2 && !strcmp (argv[1], "--help"))
    {
      puts ("\
Floppy-Disk Mounter 1.3\t\t\t(c) 1996 Martin Mares\n\
\n\
Usage: flop [-u] [-d] [<drive number or device name> [=<fs type>] ...]\n\
");
      exit (0);
    }
  checkuser ();
  if (stat ("/", &rootstat))
    err ("Cannot stat() root");
  if (argc > 1 && !strcmp (argv[1], "-u"))
    {
      argc--;
      argv++;
      unmount = 1;
    }
  if (argc > 1 && !strcmp (argv[1], "-d"))
	{
	  argc--;
	  argv++;
	  autodetect = 0;
	}
  if (!strcmp (truncate_path (argv[0]), "unflop"))
    unmount = 1;
  forbid_breaks ();
  start_mtab ();
  if (argc != 1)
    for (i = 1; i < argc; i++)
      doarg (argv[i]);
  else
    for (i = 0; i < 4; i++)
      {
		fs[0] = 0;
		devnum (i);
		if (check_device ())
		  do_mount ();
      }
  myexit (exit_code);
}
